(* equational systems, minimal -- for use in Biegelinie
   author: Walther Neuper
   050826,
   (c) due to copyright terms
*)

theory EqSystem imports Integrate Rational Root begin

consts

  occur'_exactly'_in :: 
   "[real list, real list, 'a] => bool" ("_ from _ occur'_exactly'_in _")

  (*descriptions in the related problems*)
  solveForVars       :: "real list => toreall"
  solution           :: "bool list => toreall"

  (*the CAS-command, eg. "solveSystem [x+y=1,y=2] [x,y]"*)
  solveSystem        :: "[bool list, real list] => bool list"

  (*Script-names*)
  SolveSystemScript  :: "[bool list, real list,     bool list]  
						 => bool list"
                  ("((Script SolveSystemScript (_ _ =))// (_))" 9)

axioms 
(*stated as axioms, todo: prove as theorems
  'bdv' is a constant handled on the meta-level 
   specifically as a 'bound variable'            *)

  commute_0_equality:  "(0 = a) = (a = 0)"

  (*WN0510 see simliar rules 'isolate_' 'separate_' (by RL)
    [bdv_1,bdv_2,bdv_3,bdv_4] work also for 2 and 3 bdvs, ugly !*)
  separate_bdvs_add:   
    "[| [] from [bdv_1,bdv_2,bdv_3,bdv_4] occur_exactly_in a |] 
		      			     ==> (a + b = c) = (b = c + -1*a)"
  separate_bdvs0:
    "[| some_of [bdv_1,bdv_2,bdv_3,bdv_4] occur_in b; Not (b=!=0)  |] 
		      			     ==> (a = b) = (a + -1*b = 0)"
  separate_bdvs_add1:  
    "[| some_of [bdv_1,bdv_2,bdv_3,bdv_4] occur_in c |] 
		      			     ==> (a = b + c) = (a + -1*c = b)"
  separate_bdvs_add2:
    "[| Not (some_of [bdv_1,bdv_2,bdv_3,bdv_4] occur_in a) |] 
		      			     ==> (a + b = c) = (b = -1*a + c)"



  separate_bdvs_mult:  
    "[| [] from [bdv_1,bdv_2,bdv_3,bdv_4] occur_exactly_in a; Not (a=!=0) |] 
		      			     ==>(a * b = c) = (b = c / a)"

  (*requires rew_ord for termination, eg. ord_simplify_Integral;
    works for lists of any length, interestingly !?!*)
  order_system_NxN:     "[a,b] = [b,a]"

ML {*
val thy = @{theory};

(** eval functions **)

(*certain variables of a given list occur _all_ in a term
  args: all: ..variables, which are under consideration (eg. the bound vars)
        vs:  variables which must be in t, 
             and none of the others in all must be in t
        t: the term under consideration
 *)
fun occur_exactly_in vs all t =
    let fun occurs_in' a b = occurs_in b a
    in foldl and_ (true, map (occurs_in' t) vs)
       andalso not (foldl or_ (false, map (occurs_in' t) 
                                          (subtract op = vs all)))
    end;

(*("occur_exactly_in", ("EqSystem.occur'_exactly'_in", 
			eval_occur_exactly_in "#eval_occur_exactly_in_"))*)
fun eval_occur_exactly_in _ "EqSystem.occur'_exactly'_in"
			  (p as (Const ("EqSystem.occur'_exactly'_in",_) 
				       $ vs $ all $ t)) _ =
    if occur_exactly_in (isalist2list vs) (isalist2list all) t
    then SOME ((term2str p) ^ " = True",
	       Trueprop $ (mk_equality (p, HOLogic.true_const)))
    else SOME ((term2str p) ^ " = False",
	       Trueprop $ (mk_equality (p, HOLogic.false_const)))
  | eval_occur_exactly_in _ _ _ _ = NONE;

calclist':= 
overwritel (!calclist', 
	    [("occur_exactly_in", 
	      ("EqSystem.occur'_exactly'_in", 
	       eval_occur_exactly_in "#eval_occur_exactly_in_"))
    ]);


(** rewrite order 'ord_simplify_System' **)

(* order wrt. several linear (i.e. without exponents) variables "c","c_2",..
   which leaves the monomials containing c, c_2,... at the end of an Integral
   and puts the c, c_2,... rightmost within a monomial.

   WN050906 this is a quick and dirty adaption of ord_make_polynomial_in,
   which was most adequate, because it uses size_of_term*)
(**)
local (*. for simplify_System .*)
(**)
open Term;  (* for type order = EQUAL | LESS | GREATER *)

fun pr_ord EQUAL = "EQUAL"
  | pr_ord LESS  = "LESS"
  | pr_ord GREATER = "GREATER";

fun dest_hd' (Const (a, T)) = (((a, 0), T), 0)
  | dest_hd' (Free (ccc, T)) =
    (case explode ccc of
	"c"::[] => ((("|||||||||||||||||||||", 0), T), 1)(*greatest string WN*)
      | "c"::"_"::_ => ((("|||||||||||||||||||||", 0), T), 1)
      | _ => (((ccc, 0), T), 1))
  | dest_hd' (Var v) = (v, 2)
  | dest_hd' (Bound i) = ((("", i), dummyT), 3)
  | dest_hd' (Abs (_, T, _)) = ((("", 0), T), 4);

fun size_of_term' (Free (ccc, _)) =
    (case explode ccc of (*WN0510 hack for the bound variables*)
	"c"::[] => 1000
      | "c"::"_"::is => 1000 * ((str2int o implode) is)
      | _ => 1)
  | size_of_term' (Abs (_,_,body)) = 1 + size_of_term' body
  | size_of_term' (f$t) = size_of_term' f  +  size_of_term' t
  | size_of_term' _ = 1;

fun term_ord' pr thy (Abs (_, T, t), Abs(_, U, u)) =       (* ~ term.ML *)
      (case term_ord' pr thy (t, u) of EQUAL => Term_Ord.typ_ord (T, U) | ord => ord)
  | term_ord' pr thy (t, u) =
      (if pr then 
	 let
	   val (f, ts) = strip_comb t and (g, us) = strip_comb u;
	   val _=tracing("t= f@ts= \""^
	      ((Syntax.string_of_term (thy2ctxt thy)) f)^"\" @ \"["^
	      (commas(map(Syntax.string_of_term (thy2ctxt thy)) ts))^"]\"");
	   val _=tracing("u= g@us= \""^
	      ((Syntax.string_of_term (thy2ctxt thy)) g)^"\" @ \"["^
	      (commas(map(Syntax.string_of_term (thy2ctxt thy)) us))^"]\"");
	   val _=tracing("size_of_term(t,u)= ("^
	      (string_of_int(size_of_term' t))^", "^
	      (string_of_int(size_of_term' u))^")");
	   val _=tracing("hd_ord(f,g)      = "^((pr_ord o hd_ord)(f,g)));
	   val _=tracing("terms_ord(ts,us) = "^
			   ((pr_ord o terms_ord str false)(ts,us)));
	   val _=tracing("-------");
	 in () end
       else ();
	 case int_ord (size_of_term' t, size_of_term' u) of
	   EQUAL =>
	     let val (f, ts) = strip_comb t and (g, us) = strip_comb u in
	       (case hd_ord (f, g) of EQUAL => (terms_ord str pr) (ts, us) 
	     | ord => ord)
	     end
	 | ord => ord)
and hd_ord (f, g) =                                        (* ~ term.ML *)
  prod_ord (prod_ord Term_Ord.indexname_ord Term_Ord.typ_ord) int_ord (dest_hd' f, 
						     dest_hd' g)
and terms_ord str pr (ts, us) = 
    list_ord (term_ord' pr (assoc_thy "Isac"))(ts, us);
(**)
in
(**)
(*WN0510 for preliminary use in eval_order_system, see case-study mat-eng.tex
fun ord_simplify_System_rev (pr:bool) thy subst tu = 
    (term_ord' pr thy (Library.swap tu) = LESS);*)

(*for the rls's*)
fun ord_simplify_System (pr:bool) thy subst tu = 
    (term_ord' pr thy tu = LESS);
(**)
end;
(**)
rew_ord' := overwritel (!rew_ord',
[("ord_simplify_System", ord_simplify_System false thy)
 ]);
*}
ML {*
(** rulesets **)

(*.adapted from 'order_add_mult_in' by just replacing the rew_ord.*)
val order_add_mult_System = 
  Rls{id = "order_add_mult_System", preconds = [], 
      rew_ord = ("ord_simplify_System",
		 ord_simplify_System false (theory "Integrate")),
      erls = e_rls,srls = Erls, calc = [],
      rules = [Thm ("real_mult_commute",num_str @{thm real_mult_commute}),
	       (* z * w = w * z *)
	       Thm ("real_mult_left_commute",num_str @{thm real_mult_left_commute}),
	       (*z1.0 * (z2.0 * z3.0) = z2.0 * (z1.0 * z3.0)*)
	       Thm ("real_mult_assoc",num_str @{thm real_mult_assoc}),		
	       (*z1.0 * z2.0 * z3.0 = z1.0 * (z2.0 * z3.0)*)
	       Thm ("add_commute",num_str @{thm add_commute}),	
	       (*z + w = w + z*)
	       Thm ("add_left_commute",num_str @{thm add_left_commute}),
	       (*x + (y + z) = y + (x + z)*)
	       Thm ("add_assoc",num_str @{thm add_assoc})	               
	       (*z1.0 + z2.0 + z3.0 = z1.0 + (z2.0 + z3.0)*)
	       ], 
      scr = EmptyScr}:rls;
*}
ML {*
(*.adapted from 'norm_Rational' by
  #1 using 'ord_simplify_System' in 'order_add_mult_System'
  #2 NOT using common_nominator_p                          .*)
val norm_System_noadd_fractions = 
  Rls {id = "norm_System_noadd_fractions", preconds = [], 
       rew_ord = ("dummy_ord",dummy_ord), 
       erls = norm_rat_erls, srls = Erls, calc = [],
       rules = [(*sequence given by operator precedence*)
		Rls_ discard_minus,
		Rls_ powers,
		Rls_ rat_mult_divide,
		Rls_ expand,
		Rls_ reduce_0_1_2,
		Rls_ (*order_add_mult #1*) order_add_mult_System,
		Rls_ collect_numerals,
		(*Rls_ add_fractions_p, #2*)
		Rls_ cancel_p
		],
       scr = Script ((term_of o the o (parse thy)) 
			 "empty_script")
       }:rls;
*}
ML {*
(*.adapted from 'norm_Rational' by
  *1* using 'ord_simplify_System' in 'order_add_mult_System'.*)
val norm_System = 
  Rls {id = "norm_System", preconds = [], 
       rew_ord = ("dummy_ord",dummy_ord), 
       erls = norm_rat_erls, srls = Erls, calc = [],
       rules = [(*sequence given by operator precedence*)
		Rls_ discard_minus,
		Rls_ powers,
		Rls_ rat_mult_divide,
		Rls_ expand,
		Rls_ reduce_0_1_2,
		Rls_ (*order_add_mult *1*) order_add_mult_System,
		Rls_ collect_numerals,
		Rls_ add_fractions_p,
		Rls_ cancel_p
		],
       scr = Script ((term_of o the o (parse thy)) 
			 "empty_script")
       }:rls;
*}
ML {*
(*.simplify an equational system BEFORE solving it such that parentheses are
   ( ((u0*v0)*w0) + ( ((u1*v1)*w1) * c + ... +((u4*v4)*w4) * c_4 ) )
ATTENTION: works ONLY for bound variables c, c_1, c_2, c_3, c_4 :ATTENTION
   This is a copy from 'make_ratpoly_in' with respective reductions:
   *0* expand the term, ie. distribute * and / over +
   *1* ord_simplify_System instead of termlessI
   *2* no add_fractions_p (= common_nominator_p_rls !)
   *3* discard_parentheses only for (.*(.*.))
   analoguous to simplify_Integral                                       .*)
val simplify_System_parenthesized = 
  Seq {id = "simplify_System_parenthesized", preconds = []:term list, 
       rew_ord = ("dummy_ord", dummy_ord),
      erls = Atools_erls, srls = Erls, calc = [],
      rules = [Thm ("left_distrib",num_str @{thm left_distrib}),
 	       (*"(?z1.0 + ?z2.0) * ?w = ?z1.0 * ?w + ?z2.0 * ?w"*)
	       Thm ("add_divide_distrib",num_str @{thm add_divide_distrib}),
 	       (*"(?x + ?y) / ?z = ?x / ?z + ?y / ?z"*)
	       (*^^^^^ *0* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^*)
	       Rls_ norm_Rational_noadd_fractions(**2**),
	       Rls_ (*order_add_mult_in*) norm_System_noadd_fractions (**1**),
	       Thm ("sym_real_mult_assoc",
                     num_str (@{thm real_mult_assoc} RS @{thm sym}))
	       (*Rls_ discard_parentheses *3**),
	       Rls_ collect_bdv, (*from make_polynomial_in WN051031 welldone?*)
	       Rls_ separate_bdv2,
	       Calc ("Rings.inverse_class.divide"  ,eval_cancel "#divide_e")
	       ],
      scr = EmptyScr}:rls;      
*}
ML {*
(*.simplify an equational system AFTER solving it;
   This is a copy of 'make_ratpoly_in' with the differences
   *1* ord_simplify_System instead of termlessI           .*)
(*TODO.WN051031 ^^^^^^^^^^ should be in EACH rls contained *)
val simplify_System = 
  Seq {id = "simplify_System", preconds = []:term list, 
       rew_ord = ("dummy_ord", dummy_ord),
      erls = Atools_erls, srls = Erls, calc = [],
      rules = [Rls_ norm_Rational,
	       Rls_ (*order_add_mult_in*) norm_System (**1**),
	       Rls_ discard_parentheses,
	       Rls_ collect_bdv, (*from make_polynomial_in WN051031 welldone?*)
	       Rls_ separate_bdv2,
	       Calc ("Rings.inverse_class.divide"  ,eval_cancel "#divide_e")
	       ],
      scr = EmptyScr}:rls;      
(*
val simplify_System = 
    append_rls "simplify_System" simplify_System_parenthesized
	       [Thm ("sym_add_assoc",
                      num_str (@{thm add_assoc} RS @{thm sym}))];
*)
*}
ML {*
val isolate_bdvs = 
    Rls {id="isolate_bdvs", preconds = [], 
	 rew_ord = ("e_rew_ord", e_rew_ord), 
	 erls = append_rls "erls_isolate_bdvs" e_rls 
			   [(Calc ("EqSystem.occur'_exactly'_in", 
				   eval_occur_exactly_in 
				       "#eval_occur_exactly_in_"))
			    ], 
			   srls = Erls, calc = [],
	      rules = 
             [Thm ("commute_0_equality", num_str @{thm commute_0_equality}),
	      Thm ("separate_bdvs_add", num_str @{thm separate_bdvs_add}),
	      Thm ("separate_bdvs_mult", num_str @{thm separate_bdvs_mult})],
	      scr = EmptyScr};
*}
ML {*
val isolate_bdvs_4x4 = 
    Rls {id="isolate_bdvs_4x4", preconds = [], 
	 rew_ord = ("e_rew_ord", e_rew_ord), 
	 erls = append_rls 
		    "erls_isolate_bdvs_4x4" e_rls 
		    [Calc ("EqSystem.occur'_exactly'_in", 
			   eval_occur_exactly_in "#eval_occur_exactly_in_"),
		     Calc ("Atools.ident",eval_ident "#ident_"),
		     Calc ("Atools.some'_occur'_in", 
			   eval_some_occur_in "#some_occur_in_"),
                     Thm ("not_true",num_str @{thm not_true}),
		     Thm ("not_false",num_str @{thm not_false})
			    ], 
	 srls = Erls, calc = [],
	 rules = [Thm ("commute_0_equality", num_str @{thm commute_0_equality}),
		  Thm ("separate_bdvs0", num_str @{thm separate_bdvs0}),
		  Thm ("separate_bdvs_add1", num_str @{thm separate_bdvs_add1}),
		  Thm ("separate_bdvs_add1", num_str @{thm separate_bdvs_add2}),
		  Thm ("separate_bdvs_mult", num_str @{thm separate_bdvs_mult})
                 ], scr = EmptyScr};

*}
ML {*

(*.order the equations in a system such, that a triangular system (if any)
   appears as [..c_4 = .., ..., ..., ..c_1 + ..c_2 + ..c_3 ..c_4 = ..].*)
val order_system = 
    Rls {id="order_system", preconds = [], 
	 rew_ord = ("ord_simplify_System", 
		    ord_simplify_System false thy), 
	 erls = Erls, srls = Erls, calc = [],
	 rules = [Thm ("order_system_NxN", num_str @{thm order_system_NxN})
		  ],
	 scr = EmptyScr};

val prls_triangular = 
    Rls {id="prls_triangular", preconds = [], 
	 rew_ord = ("e_rew_ord", e_rew_ord), 
	 erls = Rls {id="erls_prls_triangular", preconds = [], 
		     rew_ord = ("e_rew_ord", e_rew_ord), 
		     erls = Erls, srls = Erls, calc = [],
		     rules = [(*for precond NTH_CONS ...*)
			      Calc ("op <",eval_equ "#less_"),
			      Calc ("Groups.plus_class.plus", eval_binop "#add_")
			      (*immediately repeated rewrite pushes
					    '+' into precondition !*)
			      ],
		     scr = EmptyScr}, 
	 srls = Erls, calc = [],
	 rules = [Thm ("NTH_CONS",num_str @{thm NTH_CONS}),
		  Calc ("Groups.plus_class.plus", eval_binop "#add_"),
		  Thm ("NTH_NIL",num_str @{thm NTH_NIL}),
		  Thm ("tl_Cons",num_str @{thm tl_Cons}),
		  Thm ("tl_Nil",num_str @{thm tl_Nil}),
		  Calc ("EqSystem.occur'_exactly'_in", 
			eval_occur_exactly_in 
			    "#eval_occur_exactly_in_")
		  ],
	 scr = EmptyScr};
*}
ML {*

(*WN060914 quickly created for 4x4; 
 more similarity to prls_triangular desirable*)
val prls_triangular4 = 
    Rls {id="prls_triangular4", preconds = [], 
	 rew_ord = ("e_rew_ord", e_rew_ord), 
	 erls = Rls {id="erls_prls_triangular4", preconds = [], 
		     rew_ord = ("e_rew_ord", e_rew_ord), 
		     erls = Erls, srls = Erls, calc = [],
		     rules = [(*for precond NTH_CONS ...*)
			      Calc ("op <",eval_equ "#less_"),
			      Calc ("Groups.plus_class.plus", eval_binop "#add_")
			      (*immediately repeated rewrite pushes
					    '+' into precondition !*)
			      ],
		     scr = EmptyScr}, 
	 srls = Erls, calc = [],
	 rules = [Thm ("NTH_CONS",num_str @{thm NTH_CONS}),
		  Calc ("Groups.plus_class.plus", eval_binop "#add_"),
		  Thm ("NTH_NIL",num_str @{thm NTH_NIL}),
		  Thm ("tl_Cons",num_str @{thm tl_Cons}),
		  Thm ("tl_Nil",num_str @{thm tl_Nil}),
		  Calc ("EqSystem.occur'_exactly'_in", 
			eval_occur_exactly_in 
			    "#eval_occur_exactly_in_")
		  ],
	 scr = EmptyScr};
*}
ML {*

ruleset' := 
overwritelthy @{theory} 
	      (!ruleset', 
[("simplify_System_parenthesized", prep_rls simplify_System_parenthesized),
 ("simplify_System", prep_rls simplify_System),
 ("isolate_bdvs", prep_rls isolate_bdvs),
 ("isolate_bdvs_4x4", prep_rls isolate_bdvs_4x4),
 ("order_system", prep_rls order_system),
 ("order_add_mult_System", prep_rls order_add_mult_System),
 ("norm_System_noadd_fractions", prep_rls norm_System_noadd_fractions),
 ("norm_System", prep_rls norm_System)
 ]);
*}
ML {*


(** problems **)

store_pbt
 (prep_pbt thy "pbl_equsys" [] e_pblID
 (["system"],
  [("#Given" ,["equalities e_s", "solveForVars v_s"]),
   ("#Find"  ,["solution ss'''"](*''' is copy-named*))
  ],
  append_rls "e_rls" e_rls [(*for preds in where_*)], 
  SOME "solveSystem e_s v_s", 
  []));
store_pbt
 (prep_pbt thy "pbl_equsys_lin" [] e_pblID
 (["linear", "system"],
  [("#Given" ,["equalities e_s", "solveForVars v_s"]),
   (*TODO.WN050929 check linearity*)
   ("#Find"  ,["solution ss'''"])
  ],
  append_rls "e_rls" e_rls [(*for preds in where_*)], 
  SOME "solveSystem e_s v_s", 
  []));
store_pbt
 (prep_pbt thy "pbl_equsys_lin_2x2" [] e_pblID
 (["2x2", "linear", "system"],
  (*~~~~~~~~~~~~~~~~~~~~~~~~~*)
  [("#Given" ,["equalities e_s", "solveForVars v_s"]),
   ("#Where"  ,["LENGTH (e_s:: bool list) = 2", "LENGTH v_s = 2"]),
   ("#Find"  ,["solution ss'''"])
  ],
  append_rls "prls_2x2_linear_system" e_rls 
			     [Thm ("LENGTH_CONS",num_str @{thm LENGTH_CONS}),
			      Thm ("LENGTH_NIL",num_str @{thm LENGTH_NIL}),
			      Calc ("Groups.plus_class.plus", eval_binop "#add_"),
			      Calc ("op =",eval_equal "#equal_")
			      ], 
  SOME "solveSystem e_s v_s", 
  []));
*}
ML {*
store_pbt
 (prep_pbt thy "pbl_equsys_lin_2x2_tri" [] e_pblID
 (["triangular", "2x2", "linear", "system"],
  [("#Given" ,["equalities e_s", "solveForVars v_s"]),
   ("#Where"  ,
    ["(tl v_s) from v_s occur_exactly_in (NTH 1 (e_s::bool list))",
     "    v_s  from v_s occur_exactly_in (NTH 2 (e_s::bool list))"]),
   ("#Find"  ,["solution ss'''"])
  ],
  prls_triangular, 
  SOME "solveSystem e_s v_s", 
  [["EqSystem","top_down_substitution","2x2"]]));
store_pbt
 (prep_pbt thy "pbl_equsys_lin_2x2_norm" [] e_pblID
 (["normalize", "2x2", "linear", "system"],
  [("#Given" ,["equalities e_s", "solveForVars v_s"]),
   ("#Find"  ,["solution ss'''"])
  ],
  append_rls "e_rls" e_rls [(*for preds in where_*)], 
  SOME "solveSystem e_s v_s", 
  [["EqSystem","normalize","2x2"]]));
store_pbt
 (prep_pbt thy "pbl_equsys_lin_3x3" [] e_pblID
 (["3x3", "linear", "system"],
  (*~~~~~~~~~~~~~~~~~~~~~~~~~*)
  [("#Given" ,["equalities e_s", "solveForVars v_s"]),
   ("#Where"  ,["LENGTH (e_s:: bool list) = 3", "LENGTH v_s = 3"]),
   ("#Find"  ,["solution ss'''"])
  ],
  append_rls "prls_3x3_linear_system" e_rls 
			     [Thm ("LENGTH_CONS",num_str @{thm LENGTH_CONS}),
			      Thm ("LENGTH_NIL",num_str @{thm LENGTH_NIL}),
			      Calc ("Groups.plus_class.plus", eval_binop "#add_"),
			      Calc ("op =",eval_equal "#equal_")
			      ], 
  SOME "solveSystem e_s v_s", 
  []));
store_pbt
 (prep_pbt thy "pbl_equsys_lin_4x4" [] e_pblID
 (["4x4", "linear", "system"],
  (*~~~~~~~~~~~~~~~~~~~~~~~~~*)
  [("#Given" ,["equalities e_s", "solveForVars v_s"]),
   ("#Where"  ,["LENGTH (e_s:: bool list) = 4", "LENGTH v_s = 4"]),
   ("#Find"  ,["solution ss'''"])
  ],
  append_rls "prls_4x4_linear_system" e_rls 
			     [Thm ("LENGTH_CONS",num_str @{thm LENGTH_CONS}),
			      Thm ("LENGTH_NIL",num_str @{thm LENGTH_NIL}),
			      Calc ("Groups.plus_class.plus", eval_binop "#add_"),
			      Calc ("op =",eval_equal "#equal_")
			      ], 
  SOME "solveSystem e_s v_s", 
  []));
*}
ML {*
store_pbt
 (prep_pbt thy "pbl_equsys_lin_4x4_tri" [] e_pblID
 (["triangular", "4x4", "linear", "system"],
  [("#Given" ,["equalities e_s", "solveForVars v_s"]),
   ("#Where" , (*accepts missing variables up to diagional form*)
    ["(NTH 1 (v_s::real list)) occurs_in (NTH 1 (e_s::bool list))",
     "(NTH 2 (v_s::real list)) occurs_in (NTH 2 (e_s::bool list))",
     "(NTH 3 (v_s::real list)) occurs_in (NTH 3 (e_s::bool list))",
     "(NTH 4 (v_s::real list)) occurs_in (NTH 4 (e_s::bool list))"
     ]),
   ("#Find"  ,["solution ss'''"])
  ],
  append_rls "prls_tri_4x4_lin_sys" prls_triangular
	     [Calc ("Atools.occurs'_in",eval_occurs_in "")], 
  SOME "solveSystem e_s v_s", 
  [["EqSystem","top_down_substitution","4x4"]]));
*}
ML {*
store_pbt
 (prep_pbt thy "pbl_equsys_lin_4x4_norm" [] e_pblID
 (["normalize", "4x4", "linear", "system"],
  [("#Given" ,["equalities e_s", "solveForVars v_s"]),
   (*LENGTH is checked 1 level above*)
   ("#Find"  ,["solution ss'''"])
  ],
  append_rls "e_rls" e_rls [(*for preds in where_*)], 
  SOME "solveSystem e_s v_s", 
  [["EqSystem","normalize","4x4"]]));


(* show_ptyps();
   *)

*}
ML {*
(** methods **)

store_met
    (prep_met thy "met_eqsys" [] e_metID
	      (["EqSystem"],
	       [],
	       {rew_ord'="tless_true", rls' = Erls, calc = [], 
		srls = Erls, prls = Erls, crls = Erls, nrls = Erls},
	       "empty_script"
	       ));
store_met
    (prep_met thy "met_eqsys_topdown" [] e_metID
	      (["EqSystem","top_down_substitution"],
	       [],
	       {rew_ord'="tless_true", rls' = Erls, calc = [], 
		srls = Erls, prls = Erls, crls = Erls, nrls = Erls},
	       "empty_script"
	       ));
*}
ML {*
store_met
    (prep_met thy "met_eqsys_topdown_2x2" [] e_metID
	 (["EqSystem","top_down_substitution","2x2"],
	  [("#Given" ,["equalities e_s", "solveForVars v_s"]),
	   ("#Where"  ,
	    ["(tl v_s) from v_s occur_exactly_in (NTH 1 (e_s::bool list))",
	     "    v_s  from v_s occur_exactly_in (NTH 2 (e_s::bool list))"]),
	   ("#Find"  ,["solution ss'''"])
	   ],
	  {rew_ord'="ord_simplify_System", rls' = Erls, calc = [], 
	   srls = append_rls "srls_top_down_2x2" e_rls
				  [Thm ("hd_thm",num_str @{thm hd_thm}),
				   Thm ("tl_Cons",num_str @{thm tl_Cons}),
				   Thm ("tl_Nil",num_str @{thm tl_Nil})
				   ], 
	   prls = prls_triangular, crls = Erls, nrls = Erls},
"Script SolveSystemScript (e_s::bool list) (v_s::real list) =                " ^
"  (let e_1 = Take (hd e_s);                                                " ^
"       e_1 = ((Try (Rewrite_Set_Inst [(bdv_1, hd v_s),(bdv_2, hd (tl v_s))]" ^
"                                  isolate_bdvs False))     @@               " ^
"               (Try (Rewrite_Set_Inst [(bdv_1, hd v_s),(bdv_2, hd (tl v_s))]" ^
"                                  simplify_System False))) e_1;            " ^
"       e_2 = Take (hd (tl e_s));                                           " ^
"       e_2 = ((Substitute [e_1]) @@                                       " ^
"               (Try (Rewrite_Set_Inst [(bdv_1, hd v_s),(bdv_2, hd (tl v_s))]" ^
"                                  simplify_System_parenthesized False)) @@  " ^
"               (Try (Rewrite_Set_Inst [(bdv_1, hd v_s),(bdv_2, hd (tl v_s))]" ^
"                                  isolate_bdvs False))     @@               " ^
"               (Try (Rewrite_Set_Inst [(bdv_1, hd v_s),(bdv_2, hd (tl v_s))]" ^
"                                  simplify_System False))) e_2;            " ^
"       e__s = Take [e_1, e_2]                                             " ^
"   in (Try (Rewrite_Set order_system False)) e__s)"
(*---------------------------------------------------------------------------
  this script does NOT separate the equations as above, 
  but it does not yet work due to preliminary script-interpreter,
  see eqsystem.sml 'script [EqSystem,top_down_substitution,2x2] Vers.2'

"Script SolveSystemScript (e_s::bool list) (v_s::real list) =         " ^
"  (let e__s = Take e_s;                                              " ^
"       e_1 = hd e__s;                                               " ^
"       e_2 = hd (tl e__s);                                          " ^
"       e__s = [e_1, Substitute [e_1] e_2]                         " ^
"   in ((Try (Rewrite_Set_Inst [(bdv_1, hd v_s),(bdv_2, hd (tl v_s))]" ^
"                                  simplify_System_parenthesized False)) @@ " ^
"       (Try (Rewrite_Set_Inst [(bdv_1, hd v_s),(bdv_2, hd (tl v_s))] " ^
"                              isolate_bdvs False))              @@   " ^
"       (Try (Rewrite_Set_Inst [(bdv_1, hd v_s),(bdv_2, hd (tl v_s))]" ^
"                                  simplify_System False))) e__s)"
---------------------------------------------------------------------------*)
	  ));
*}
ML {*
store_met
    (prep_met thy "met_eqsys_norm" [] e_metID
	      (["EqSystem","normalize"],
	       [],
	       {rew_ord'="tless_true", rls' = Erls, calc = [], 
		srls = Erls, prls = Erls, crls = Erls, nrls = Erls},
	       "empty_script"
	       ));
*}
ML {*
store_met
    (prep_met thy "met_eqsys_norm_2x2" [] e_metID
	      (["EqSystem","normalize","2x2"],
	       [("#Given" ,["equalities e_s", "solveForVars v_s"]),
		("#Find"  ,["solution ss'''"])],
	       {rew_ord'="tless_true", rls' = Erls, calc = [], 
		srls = append_rls "srls_normalize_2x2" e_rls
				  [Thm ("hd_thm",num_str @{thm hd_thm}),
				   Thm ("tl_Cons",num_str @{thm tl_Cons}),
				   Thm ("tl_Nil",num_str @{thm tl_Nil})
				   ], 
		prls = Erls, crls = Erls, nrls = Erls},
"Script SolveSystemScript (e_s::bool list) (v_s::real list) =                " ^
"  (let e__s = ((Try (Rewrite_Set norm_Rational False)) @@                   " ^
"               (Try (Rewrite_Set_Inst [(bdv_1, hd v_s),(bdv_2, hd (tl v_s))]" ^
"                                  simplify_System_parenthesized False)) @@  " ^
"               (Try (Rewrite_Set_Inst [(bdv_1, hd v_s),(bdv_2, hd (tl v_s))]" ^
"                                                    isolate_bdvs False)) @@ " ^
"               (Try (Rewrite_Set_Inst [(bdv_1, hd v_s),(bdv_2, hd (tl v_s))]" ^
"                                  simplify_System_parenthesized False)) @@  " ^
"               (Try (Rewrite_Set order_system False))) e_s                  " ^
"   in (SubProblem (EqSystem',[linear,system],[no_met])                      " ^
"                  [BOOL_LIST e__s, REAL_LIST v_s]))"
	       ));

*}
ML {*
(*this is for NTH only*)
val srls = Rls {id="srls_normalize_4x4", 
		preconds = [], 
		rew_ord = ("termlessI",termlessI), 
		erls = append_rls "erls_in_srls_IntegrierenUnd.." e_rls
				  [(*for asm in NTH_CONS ...*)
				   Calc ("op <",eval_equ "#less_"),
				   (*2nd NTH_CONS pushes n+-1 into asms*)
				   Calc("Groups.plus_class.plus", eval_binop "#add_")
				   ], 
		srls = Erls, calc = [],
		rules = [Thm ("NTH_CONS",num_str @{thm NTH_CONS}),
			 Calc("Groups.plus_class.plus", eval_binop "#add_"),
			 Thm ("NTH_NIL",num_str @{thm NTH_NIL})],
		scr = EmptyScr};
store_met
    (prep_met thy "met_eqsys_norm_4x4" [] e_metID
	      (["EqSystem","normalize","4x4"],
	       [("#Given" ,["equalities e_s", "solveForVars v_s"]),
		("#Find"  ,["solution ss'''"])],
	       {rew_ord'="tless_true", rls' = Erls, calc = [], 
		srls = append_rls "srls_normalize_4x4" srls
				  [Thm ("hd_thm",num_str @{thm hd_thm}),
				   Thm ("tl_Cons",num_str @{thm tl_Cons}),
				   Thm ("tl_Nil",num_str @{thm tl_Nil})
				   ], 
		prls = Erls, crls = Erls, nrls = Erls},
(*GOON met ["EqSystem","normalize","4x4"] @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@*)
"Script SolveSystemScript (e_s::bool list) (v_s::real list) =                " ^
"  (let e__s =                                                               " ^
"     ((Try (Rewrite_Set norm_Rational False)) @@                            " ^
"      (Repeat (Rewrite commute_0_equality False)) @@                        " ^
"      (Try (Rewrite_Set_Inst [(bdv_1, NTH 1 v_s),(bdv_2, NTH 2 v_s ),     " ^
"                              (bdv_3, NTH 3 v_s),(bdv_3, NTH 4 v_s )]     " ^
"                             simplify_System_parenthesized False))    @@    " ^
"      (Try (Rewrite_Set_Inst [(bdv_1, NTH 1 v_s),(bdv_2, NTH 2 v_s ),     " ^
"                              (bdv_3, NTH 3 v_s),(bdv_3, NTH 4 v_s )]     " ^
"                             isolate_bdvs_4x4 False))                 @@    " ^
"      (Try (Rewrite_Set_Inst [(bdv_1, NTH 1 v_s),(bdv_2, NTH 2 v_s ),     " ^
"                              (bdv_3, NTH 3 v_s),(bdv_3, NTH 4 v_s )]     " ^
"                             simplify_System_parenthesized False))    @@    " ^
"      (Try (Rewrite_Set order_system False)))                           e_s " ^
"   in (SubProblem (EqSystem',[linear,system],[no_met])                      " ^
"                  [BOOL_LIST e__s, REAL_LIST v_s]))"
));
store_met
(prep_met thy "met_eqsys_topdown_4x4" [] e_metID
	  (["EqSystem","top_down_substitution","4x4"],
	   [("#Given" ,["equalities e_s", "solveForVars v_s"]),
	    ("#Where" , (*accepts missing variables up to diagonal form*)
	     ["(NTH 1 (v_s::real list)) occurs_in (NTH 1 (e_s::bool list))",
	      "(NTH 2 (v_s::real list)) occurs_in (NTH 2 (e_s::bool list))",
	      "(NTH 3 (v_s::real list)) occurs_in (NTH 3 (e_s::bool list))",
	      "(NTH 4 (v_s::real list)) occurs_in (NTH 4 (e_s::bool list))"
	      ]),
	    ("#Find"  ,["solution ss'''"])
	    ],
	   {rew_ord'="ord_simplify_System", rls' = Erls, calc = [], 
	    srls = append_rls "srls_top_down_4x4" srls [], 
	    prls = append_rls "prls_tri_4x4_lin_sys" prls_triangular
			      [Calc ("Atools.occurs'_in",eval_occurs_in "")], 
	    crls = Erls, nrls = Erls},
(*FIXXXXME.WN060916: this script works ONLY for exp 7.79 @@@@@@@@@@@@@@@@@@@@*)
"Script SolveSystemScript (e_s::bool list) (v_s::real list) =                " ^
"  (let e_1 = NTH 1 e_s;                                                    " ^
"       e_2 = Take (NTH 2 e_s);                                             " ^
"       e_2 = ((Substitute [e_1]) @@                                         " ^
"              (Try (Rewrite_Set_Inst [(bdv_1,NTH 1 v_s),(bdv_2,NTH 2 v_s)," ^
"                                      (bdv_3,NTH 3 v_s),(bdv_4,NTH 4 v_s)]" ^
"                                 simplify_System_parenthesized False)) @@   " ^
"              (Try (Rewrite_Set_Inst [(bdv_1,NTH 1 v_s),(bdv_2,NTH 2 v_s)," ^
"                                      (bdv_3,NTH 3 v_s),(bdv_4,NTH 4 v_s)]" ^
"                                 isolate_bdvs False))                  @@   " ^
"              (Try (Rewrite_Set_Inst [(bdv_1,NTH 1 v_s),(bdv_2,NTH 2 v_s)," ^
"                                      (bdv_3,NTH 3 v_s),(bdv_4,NTH 4 v_s)]" ^
"                                 norm_Rational False)))             e_2     " ^
"    in [e_1, e_2, NTH 3 e_s, NTH 4 e_s])"
));
*}

end